%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% TEMPORAL DEBLURRING SIMULATION
%
% This simulation program show a simple temporal deblurring example of a
% video of the globally shigted house image. This simulation program consists of
% four steps:
%
%   1. LOAD A TEST IMAGE
%      We use a pepper image as a test image in this simulation.
%
%   2. CREATE A SIMULATED GROUND-TRUTH SEQUENCE
%      We create a simulated ground-truth sequence by shifting the house
%      image.
%
%   3. BLUR THE SIMULATED SEQUENCE WITH TEMPORAL PSF AND ADD SOME NOISE
%      The temporal PSF in this example is a 1x1x8 uniform PSF. Using the
%      PSF, we blur the simulated sequence and add weak noise (white
%      Gaussian noise with standard deviation 2). Assuming that the blurred
%      noisy sequence is the output of the space-time interpolator, the
%      next step is deblurring.
%
%   4. DEBLUR THE BLURRED NOISY SEQUENCE
%      The final step is deblurring. We use the true PSF in this example.
%      Using the steepest descent method, we iteratively estimate the
%      underlying sequence with 3-D (digital) total valiation.
%
%   5. DEBLUR THE BLURRED NOISY SEQUENCE IN THE SPACE DOMAIN
%      To compare our space-time debluring to the 2-D motion deblurring
%      approach in the space domain, we deblur one motion blurred frame 
%      with the true 2-D motion PSF. Using the steepest descent method, we
%      iteratively estimate the underlying sequence with 2-D (digital)
%      total valiation.
%
% Note : An appropriate mirroring around the edges of the 3-D array of the
% videos is also necessary to avoid ringing. To simplify the simulation
% code, the mirroring is omitted from this code.
%
% wrote by Hiroyuki Takeda, Feb 9th, 2011


%%%%%%%%%%%%%%%%%%%%%%%%
% 1. LOAD A TEST IMAGE %
%%%%%%%%%%%%%%%%%%%%%%%%
house = imread('house.tiff');
ycc = rgb2ycbcr(house);
img = double(ycc(:,:,1)); % use only the luminance channel
[N M] = size(img);

% this image for obtaining the ground truth PSF
imp = zeros(N, M);
imp(N/2, M/2) = 1;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 2. CREATE A GROUND-TRUTH SEQUENCE %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
th = 0 : 120;
radius = 6;
% generate motion shift vectors
motion = [round(real(radius * exp(1i * th(:) / (2 * pi)))),...
          round(imag(radius * exp(1i * th(:) / (2 * pi))))];
seq_true = zeros(N-radius*2, M-radius*2, length(th));
seq_true_imp = zeros(N-radius*2, M-radius*2, length(th));
for i = 1 : length(th)
    % global shift (pure translational motion)
    seq_true(:,:,i) = img(1+radius+motion(i,1):end-radius+motion(i,1),...
                          1+radius+motion(i,2):end-radius+motion(i,2));
    % global shift of the impulse image
    seq_true_imp(:,:,i) = imp(1+radius+motion(i,1):end-radius+motion(i,1),...
                              1+radius+motion(i,2):end-radius+motion(i,2));
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 3. BLUR THE SEQUENCE WITH A TEMPORAL PSF %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% temporal PSF (1x1x8 uniform PSF)
G = ones(1, 1, 9);
G(:,:,9) = 0;
G = G / sum(G(:));

% convolve the shifthing sequence with the temporal PSF 
seq_blur = convn(seq_true, G, 'same');
seq_true = seq_true(:,:,5:end-4);
seq_blur = seq_blur(:,:,5:end-4);

% generate the 2-D ground truth PSF
truePSFs = convn(seq_true_imp, G, 'same');
truePSFs = truePSFs(:,:,5:end-4);
motion = motion(5:end-4,:);

% add some noise
sigma = 2;
randn('state', 0);
seq_blur = seq_blur + randn(size(seq_blur)) * sigma;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
% 4. DEBLURRING IN THE SPACE-TIME DOMAIN WITH THE TRUE 3-D PSF AND TOTAL VARIATION %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
IT = 400;           % the number of iterations of the steepest descent
step = 0.5;         % the step size of the steepest descent update
lambda = 0.15;      % the regularization parameter
Gt = flipdim(G, 3); % the flipped version of the temporal PSF
u = seq_blur;       % initialize with the blurred sequence

for it = 1 : IT
   
    % the gradient from the likelihood term
    ec = convn(seq_blur - convn(u, G, 'same'), Gt, 'same'); % G^T (y - G x)
    
    % the gradient from the TV regularization term
    % x1-direction
    tmp = sign(u - ishift3(u, 1, 0, 0));
    tv = tmp - ishift3(tmp, -1, 0, 0);
    % x2-direction
    tmp = sign(u - ishift3(u, 0, 1, 0));
    tv = tv + tmp - ishift3(tmp, 0, -1, 0);
    % t-direction
    tmp = sign(u - ishift3(u, 0, 0, 1));
    tv = tv + tmp - ishift3(tmp, 0, 0, -1);
    
    % update
    update = step * (ec - lambda * tv);
    update(:,:,1:9) = 0;
    update(:,:,end-8:end) = 0;
    u = u + update;
    
    % display the estimation status
    figure(1);
    fnum = 40;
    subplot(1,3,1); imagesc(u(:,:,fnum)); colormap(gray); axis image; colorbar;
    title(sprintf('The current estimate of the frame at t=%d, %d iterations', fnum, it));
    subplot(1,3,2); imagesc(ec(:,:,fnum)); colormap(gray); axis image; colorbar;
    title('The gradient of the likelihood term');
    subplot(1,3,3); imagesc(tv(:,:,fnum)); colormap(gray); axis image; colorbar;
    title('The gradient of the total valiation term');
    pause(0.1);
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
% 5. DEBLURRING IN THE SPATIAL DOMAIN WITH THE TRUE 2-D PSF AND TOTAL VARIATION %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
IT = 400;           % the number of iterations of the steepest descent
step = 0.5;         % the step size of the steepest descent update
lambda = 0.35;      % the regularization parameter
frame = 40;
G2D = truePSFs(122-6-motion(frame,1):122+6-motion(frame,1),122-6-motion(frame,2):122+6-motion(frame,2),frame);
G2Dt = flipdim(flipdim(G2D, 1), 2); % the flipped version of the temporal PSF
u2D = seq_blur(:,:,frame);       % initialize with the blurred sequence

for it = 1 : IT
   
    % the gradient from the likelihood term
    ec = conv2(seq_blur(:,:,frame) - conv2(u2D, G2D, 'same'), G2Dt, 'same'); % G^T (y - G x)
    
    % the gradient from the TV regularization term
    % x1-direction
    tmp = sign(u2D - ishift3(u2D, 1, 0, 0));
    tv = tmp - ishift3(tmp, -1, 0, 0);
    % x2-direction
    tmp = sign(u2D - ishift3(u2D, 0, 1, 0));
    tv = tv + tmp - ishift3(tmp, 0, -1, 0);
    
    % update
    update = step * (ec - lambda * tv);
    update(1:6,:) = 0; update(end-5:end,:) = 0;
    update(:,1:6) = 0; update(:,end-5:end) = 0;
    u2D = u2D + update;
    
    % display the estimation status
    figure(1);
    subplot(1,3,1); imagesc(u2D); colormap(gray); axis image; colorbar;
    title(sprintf('The current estimate of the frame at t=%d, %d iterations', frame, it));
    subplot(1,3,2); imagesc(ec(6:end-5,6:end-5)); colormap(gray); axis image; colorbar;
    title('The gradient of the likelihood term');
    subplot(1,3,3); imagesc(tv); colormap(gray); axis image; colorbar;
    title('The gradient of the total valiation term');
    pause(0.1);
end

% display the result
figure(2);
fnum = 40;
subplot(2,2,1); imagesc(seq_true(:,:,fnum)); colormap(gray); axis image; colorbar;
title(sprintf('The original frame at t = %d', fnum));
subplot(2,2,2); imagesc(seq_blur(:,:,fnum)); colormap(gray); axis image; colorbar;
error = seq_true(:,:,fnum) - seq_blur(:,:,fnum);
psnr = 20 * log10(255 / sqrt(mean(error(:).^2)));
title(sprintf('The simulated blurly noisy frame at t = %d, PSNR = %4.2f [dB]', fnum, psnr));
subplot(2,2,3); imagesc(u2D); colormap(gray); axis image; colorbar;
error = seq_true(:,:,fnum) - u2D;
psnr = 20 * log10(255 / sqrt(mean(error(:).^2)));
title(sprintf('The deblurred frame at t = %d by the 2-D deblurring method, PSNR = %4.2f [dB]', fnum, psnr));
subplot(2,2,4); imagesc(u(:,:,fnum)); colormap(gray); axis image; colorbar;
error = seq_true(:,:,fnum) - u(:,:,fnum);
psnr = 20 * log10(255 / sqrt(mean(error(:).^2)));
title(sprintf('The deblurred frame at t = %d by our 3-D method, PSNR = %4.2f [dB]', fnum, psnr));



